# TapeArchive, a tar file module for Ruby
# Copyright (C) 2003  Jamis Buck (jgb3@email.byu.edu)
#
# TapeArchive is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# TapeArchive is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TapeArchive; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# =========================================================================

require 'parmvalid'
require 'stream'

# TapeArchive is a module that allows Ruby programs to create and access 'tar'
# files,  natively.  It depends on no external utilities (ie, the 'tar'
# program), so you can use it on systems where 'tar' is not normally installed.
#
# To use it to list all entries in a tar file:
#
#   require 'tar'
#
#   TapeArchive::File.open( "some.tar" ) do |file|
#     file.each_entry do |entry|
#       puts "#{entry.name} :: #{entry.size} bytes"
#     end
#   end
#
# To actually extract the contents of each entry in a tar file:
#
#   TapeArchive::File.open( "some.tar" ) do |file|
#     file.each_entry do |entry|
#       content = entry.contents
#     end
#   end
#
# Alternatively, you can get an InputStream for each entry by calling
# entry.get_input_stream.
#
# To create a new tar file:
#
#   TapeArchive::File.create( "some-new.tar" ) do |file|
#     file.new_entry "somefile.txt"
#     file.puts "text"
#
#     file.new_entry "anotherfile.txt"
#     file.puts "more text"
#   end
#
# To append to the end of an existing tar file (or create a new file):
#
#   TapeArchive::File.append( "some-old.tar" ) do |file|
#     file.new_entry "appended-file.txt"
#     file.puts "text"
#   end
module TapeArchive

  # This exception gets raised if a file that is not a tar file is read.
  class BadTapeArchiveException < RuntimeError; end

  # This class represents the header record for a single tar entry.  It is also
  # used to represent an "empty block", which pads the end of a tar file.  This
  # class should almost never be instantiated directly.
  class RecordHeader
    include ParameterValidator

    # the size of a record
    RECORD_SIZE = 512

    NAME_SIZE   = 100
    UNAME_LEN   = 32
    GNAME_LEN   = 32
    PREFIX_LEN  = 155

    attr_accessor :name
    attr_accessor :mode
    attr_accessor :uid
    attr_accessor :gid
    attr_accessor :size
    attr_accessor :mtime
    attr_accessor :chksum
    attr_accessor :linkflag
    attr_accessor :linkname
    attr_accessor :magic
    attr_accessor :version
    attr_accessor :uname
    attr_accessor :gname
    attr_accessor :devmajor
    attr_accessor :devminor
    attr_accessor :prefix

    # Instantiate an empty header record, composed entirely of null bytes.
    def RecordHeader.make_nil
      new( :name=>"", :mode=>0, :uid=>0, :gid=>0, :size=>0, :mtime=>0, :chksum=>0,
           :linkflag=>"", :magic=>"", :version=>0, :uname=>"", :gname=>"",
           :devmajor=>0, :devminor=>0, :prefix=>"" )
    end

    # Read the record header from the given stream and instantiate it.
    def RecordHeader.from_stream( stream )
      fields = stream.read( RECORD_SIZE ).unpack( "A#{NAME_SIZE}" + # record name
                                                  "A8A8A8" +        # mode, uid, gid
                                                  "A12A12A8" +      # size, mtime, chksum
                                                  "A" +             # linkflag
                                                  "A#{NAME_SIZE}" + # link name
                                                  "A6A2" +          # magic, version
                                                  "A#{UNAME_LEN}" + # uname
                                                  "A#{GNAME_LEN}" + # gname
                                                  "A8A8" +          # dev major, dev minor
                                                  "A#{PREFIX_LEN}"  # prefix
                                                )

      name = fields.shift
      mode = fields.shift.oct
      uid = fields.shift.oct
      gid = fields.shift.oct
      size = fields.shift.oct
      mtime = fields.shift.oct
      chksum = fields.shift.oct
      linkflag = fields.shift
      linkname = fields.shift
      magic = fields.shift
      version = fields.shift.oct
      uname = fields.shift
      gname = fields.shift
      devmajor = fields.shift.oct
      devminor = fields.shift.oct
      prefix = fields.shift

      new( :name=>name, :mode=>mode, :uid=>uid, :gid=>gid, :size=>size, :mtime=>mtime, :chksum=>chksum,
           :linkflag=>linkflag, :magic=>magic, :version=>version, :uname=>uname, :gname=>gname,
           :devmajor=>devmajor, :devminor=>devminor, :prefix=>prefix )
    end

    VALID_INIT_PARMS = { :name=>REQUIRED, :mode=>0600, :uid=>1000, :gid=>100, :size=>0,
                         :mtime=>Time.now.to_i, :chksum=>0, :linkflag=>"0", :magic=>"ustar",
                         :version=>0, :uname=>"", :gname=>"", :devmajor=>0, :devminor=>0,
                         :prefix=>"" }

    # Instantiate a new record header from the given keyword arguments.
    def initialize( args={} )
      validate_and_instantiate_parameters( args, VALID_INIT_PARMS )

      @mtime = Time.at( @mtime )
      @name = @prefix + @name
      @mode |= 0100000

      return if nil?

      raise BadTapeArchiveException, "bad magic value (#{@magic})" if @magic != "ustar"
    end

    # Query whether this header is nil.
    def nil?
      @name.length < 1
    end

    # Convert the header to its string value, which will be written to a file.  If checksumming
    # is true, then the chksum field will be left blank so that a checksum can be computed.
    def to_s( checksumming=false )
      if nil?
        return "\0" * RECORD_SIZE
      else
        [ @name, to_oct( @mode, 7 ), to_oct( @uid, 7 ), to_oct( @gid, 7 ), to_oct( @size, 11 ),
          to_oct( @mtime.to_i, 11 ), ( checksumming ? " " * 7 : to_oct( @chksum, 6 ) ), " ", @linkflag,
          @linkname, @magic, "  ", @uname, @gname,
          ( @devmajor == 0 ? "" : to_oct( @devmajor, 7 ) ),
          ( @devmajor == 0 ? "" : to_oct( @devminor, 7 ) ),
          @prefix ] .
        pack( "a#{NAME_SIZE}a8a8a8a12a12a7aaa#{NAME_SIZE}a5a3a#{UNAME_LEN}a#{GNAME_LEN}a8a8a#{PREFIX_LEN}" )
      end
    end

    # Compute the checksum for this header.
    def compute_checksum
      sum = 0
      to_s( true ).each_byte { |b| sum += b }
      return sum
    end

    # A convenience function for converting a number to octal, with the given width (padded with 0's).
    def to_oct( n, s=0 )
      "%0#{s}o" % n
    end

    # Compute the extra padding needed to bring a block of size 'size' to RECORD_SIZE bytes.
    def RecordHeader.padding( size )
      extra = RECORD_SIZE - ( size % RECORD_SIZE )
      extra < RECORD_SIZE ? extra : 0
    end
  end

  # This class represents a single entry in a tar file.
  class Entry
    include ParameterValidator

    attr_reader :header

    # This is the class returned by Entry when an input stream is requested.  It basically
    # maps a subset of a stream and treats it as its own stream.
    class InputStream < ::InputStream

      # Instantiate a new Entry::InputStream from the given stream, using the given start
      # position and length to define the portion of the stream that this one maps.
      def initialize( file, start, length )
        @file = file
        @start = start
        @length = length
        @last_pos = @start
        @file.pos = @start
      end

      # Read <tt>num</tt> bytes from the stream.  If <tt>num</tt> is nil, read to the end
      # of the stream.
      def read( num=nil )
        sanity_checks

        remaining = ( @start + @length ) - @file.pos
        num = remaining if num.nil? or num > remaining

        return nil if num < 1

        data = @file.read( num )
        @last_pos = @file.pos

        return data
      end

      # Return the position of the stream.
      def pos
        sanity_checks
        @file.pos - @start
      end

      # Set the position of the stream.
      def pos=( num )
        sanity_checks
        new_pos = num - @start
        raise "seek beyond EOF" if new_pos > @start + @length
        @last_pos = new_pos
        @file.pos = new_pos
        num
      end

      # Attempt to close the stream.  This will throw an exception, since Entry::InputStream's
      # should never be closed directly.  Instead, close the stream that this maps to.
      def close
        raise "you cannot close an entry input stream directly"
      end

      # Query whether the input stream is closed.
      def closed?
        @file.closed?
      end

      # Rewind the stream to the start of the entry.
      def rewind
        sanity_checks
        @file.pos = @start
        @last_pos = @start
      end

      # Query whether the end of the entry has been reached.
      def eof?
        sanity_checks
        @file.pos - @start >= @length
      end

      # Do some sanity checks to determine if the integrity of the stream has
      # been compromised.
      def sanity_checks
        raise "underlying archive closed" if @file.closed?
        raise "simultaneous read detected" if @file.pos != @last_pos
      end

      private :sanity_checks
    end

    # Instantiate a brand new entity that is not associated with a specific file.
    def Entry.create( args={} )
      e = allocate
      e.instance_eval { initialize_create( args ) }
      return e
    end

    # Instantiate a new entity that begins at the current position in the given
    # file.
    def initialize( file )
      @file = file
      @header = RecordHeader.from_stream( @file )
      @position = @file.pos
      @real_size = @header.size
      
      @real_size += RecordHeader.padding( @header.size )

      file.pos += @real_size
    end

    VALID_CREATE_PARMS = { :name=>REQUIRED, :mode=>0600, :uid=>1000, :gid=>100, :uname=>"", :gname=>"" }

    # Initialize a newly created entity.
    def initialize_create( args={} )
      args = validate_parameters( args, VALID_CREATE_PARMS )

      @file = nil
      @position = nil
      @real_size = nil

      @header = RecordHeader.new( args )
    end

    private :initialize_create

    # Get the input stream for the entity.
    def get_input_stream
      raise "not a modifiable entity" if @file.nil?
      raise "entity is of a type that has no content" if not file?
      return InputStream.new( @file, @position, @header.size )
    end

    # Get the output stream for the entity.
    def get_output_stream
      @contents = "" unless @contents
      return StringOutputStream.new( @contents )
    end

    # Write the entity (with its header) to the given stream, returning the actual
    # number of blocks written.
    def write( stream )
      blocks_written = 0

      @contents ||= ""

      @header.size = @contents.length
      @header.chksum = header.compute_checksum

      hdr_bytes = @header.to_s

      stream << hdr_bytes

      extra = RecordHeader.padding( hdr_bytes.length )
      stream << "\0" * extra

      blocks_written += ( hdr_bytes.length + extra ) / RecordHeader::RECORD_SIZE

      stream << @contents

      extra = RecordHeader.padding( @contents.length )
      stream << "\0" * extra

      blocks_written += ( @contents.length + extra ) / RecordHeader::RECORD_SIZE

      return blocks_written
    end

    # Get the contents of this entity.
    def contents
      @contents = get_input_stream.read unless @contents
      @contents
    end

    # Get the name of this entity.
    def name
      @header.name
    end

    # Get the size of this entity.
    def size
      @header.size
    end

    # Is the entity nil?
    def nil?
      @header.nil?
    end

    # Does the entity represent a file?
    def file?
      @header.linkflag == "0" || @header.linkflag.length < 1
    end

    # Does the entity represent a link?
    def link?
      ( "1".."2" ).include? @header.linkflag
    end

    # Does the entity represent a character device?
    def chardev?
      @header.linkflag == "3"
    end

    # Does the entity represent a block device?
    def blockdev?
      @header.linkflag == "4"
    end

    # Does the entity represent a directory?
    def directory?
      @header.linkflag == "5"
    end

    # Does the entity represent a FIFO special file?
    def fifo?
      @header.linkflag == "6"
    end
  end
  
  # The parent class of the different archive file classes.
  class File

    # Represents a read-only tar file, which may be queried but not
    # modified.
    class ReadOnlyFile < File

      # Open an existing tar file with the given filename.  If the block is given,
      # execute the block (passing the new file to it) and then close the file.  Otherwise,
      # return the file.
      def ReadOnlyFile.open( filename, &block )
        new( filename, &block )
      end

      # Initialize a newly opened tar file.
      def initialize( filename, &block )
        super()

        @file = ::File.open( filename, "r" )
        @next_entry_at = 0

        unless block.nil?
          yield self
          close
        end
      end

      # Return the next entry in the tar file, or nil if there are no more entries.
      def next_entry
        @file.pos = @next_entry_at

        @current_entry = Entry.new( @file )
        return nil if @current_entry.nil?

        @entries[ @current_entry.name ] = @current_entry
        @next_entry_at = @file.pos

        @current_input_stream = @current_entry.get_input_stream
        return @current_entry
      end

      # Execute the attached block for each entry in the tar file.
      def each_entry
        while e = next_entry
          yield e
        end
      end

      # Read the given number of bytes from the current entry.  If num is nil, then
      # read the entire contents of the current entry.  If there is no current entry,
      # this throws an exception.
      def read( num=nil )
        raise "no current entry" if @current_entry.nil?
        @current_input_stream.read( num )
      end

      # Search the archive for the entry with the given name.  Returns nil if the
      # named entry could not be found.
      def []( entry )
        return @entries[ entry ] if @entries[ entry ]

        save_next_at = @next_entry_at

        begin
          while e = next_entry
            return e if e.name == entry
          end
        ensure
          @next_entry_at = save_next_at
        end

        return nil
      end

      # Return an array of all entries in the archive.  The array is not guaranteed to
      # list the entries in the same order as they existed in the archive.  It is safe
      # to call next_entry after this, to continue an iteration that was previously
      # active.
      def entries
        save_next_at = @next_entry_at

        begin
          loop { e = next_entry || break }
        ensure
          @next_entry_at = save_next_at
        end

        @entries.values
      end
    end

    # This class represents a tar file that may be written to, but not queried.
    class WriteOnlyFile < File
      include ParameterValidator

      # Open the given file in "write-only" mode.  If a block is given, pass the new
      # file to the block, execute the block, and then close the file.  Otherwise, return
      # the file.
      def WriteOnlyFile.create( filename, &block )
        new( filename, false, &block )
      end

      # Open the given file in "write-only" mode, for appending.  If a block is given, pass
      # the new file to the block, execute the block, and then close the file.  Otherwise,
      # return the file.
      def WriteOnlyFile.append( filename, &block )
        new( filename, true, &block )
      end

      # Initialize a new write-only tar file reference.  If 'appending' is true, then the
      # file is opened for appending and the file pointer is moved to the end of the last
      # valid block.  If the file does not exist, it is created.  If it does exist and the
      # file is not opened for append, the file is truncated.
      def initialize( filename, appending, &block )
        super()

        access = ( appending ? ( ::File.exist?( filename ) ? "r+" : "w+" ) : "w" )

        @file = ::File.open( filename, access )
        @current_entry = nil
        @total_blocks = 0

        # if we are appending, look for the last record written to the file (before the
        # final padding).
        if appending
          while not @file.eof?
            entry = Entry.new( @file )
            if entry.nil?
              @file.pos -= RecordHeader::RECORD_SIZE
              break
            else
              @entries[ entry.name ] = entry
            end
          end
          @total_blocks = @file.pos / RecordHeader::RECORD_SIZE
        end

        unless block.nil?
          yield self
          close
        end
      end

      # Execute the given block using instance_eval
      def transact( &block )
        instance_eval &block
      end

      # Create a new entry with the given filename and arguments.  Return the new
      # entry.
      def new_entry( filename, args={} )
        args[ :name ] = filename

        args = validate_parameters( args, Entry::VALID_CREATE_PARMS )

        raise "duplicate entry name '#{filename}'" if @entries[ filename ]

        @total_blocks += @current_entry.write( @file ) unless @current_entry.nil?

        @current_entry = Entry.create( args )
        @current_stream = @current_entry.get_output_stream

        @entries[ @current_entry.name ] = @current_entry
        return @current_entry
      end

      # Write the given string to the current entry.  If there is no current entry, raise an
      # exception.
      def write( s )
        raise "no current entry" if @current_entry.nil?
        @current_stream.write s
      end

      alias :print :write

      # Write the given string to the current entry, appending a newline.
      def puts( s="" )
        write( s + $/ )
      end

      # Close the current file, padding it out to a multiple of 20 blocks (the tar default).
      def close
        @total_blocks += @current_entry.write( @file ) unless @current_entry.nil?
        @current_entry = nil

        nil_block_text = RecordHeader.make_nil.to_s
        2.times { @file.write nil_block_text; @total_blocks += 1 }

        if ( extra = 20 - @total_blocks % 20 ) < 20
          extra.times { @file.write nil_block_text }
        end

        super
      end
    end

    private_class_method :new

    # Create a new read-only tar file.
    def File.open( filename, &block )
      ReadOnlyFile.open( filename, &block )
    end

    # Create a new write-only tar file.
    def File.create( filename, &block )
      WriteOnlyFile.create( filename, &block )
    end

    # Create a new write-only tar file for appending to an existing tar file.
    def File.append( filename, &block )
      WriteOnlyFile.append( filename, &block )
    end

    # Common initialization for all descendent classes.
    def initialize
      @entries = Hash.new
    end

    # Close the file.
    def close
      @file.close
    end

    # Query whether the file is closed or not.
    def closed?
      @file.closed?
    end
  end

end

